@@ -44,6 +44,11 @@
  *  • Comparing mode (#UhmServer:enable-logging: %FALSE, #UhmServer:enable-online: %TRUE): Requests are sent to the real server online, and
  *    the request–response pairs are compared against those in an existing log file to see if the log file is up-to-date.
  *
+ * Starting with version 0.11.0 hosts are automatically extracted and stored in hosts trace files. These files are
+ * used during replay so no host configuration using uhm_resolver_add_A() is necessary in code any more.
+ * Requires libsoup 3.5.1 or higher. If that version of libsoup is not available, uhm_resolver_add_A() must continue
+ * to be used.
+ *
  * Since: 0.1.0
  */
 
@@ -119,6 +124,10 @@ struct _UhmServerPrivate {
 	gboolean enable_online;
 	gboolean enable_logging;
 
+	GFile *hosts_trace_file;
+	GFileOutputStream *hosts_output_stream;
+	GHashTable *hosts;
+
 	GByteArray *comparison_message;
 	enum {
 		UNKNOWN,
@@ -327,6 +336,7 @@ static void
 uhm_server_init (UhmServer *self)
 {
 	self->priv = uhm_server_get_instance_private (self);
+	self->priv->hosts = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
 }
 
 static void
@@ -337,7 +347,9 @@ uhm_server_dispose (GObject *object)
 	g_clear_object (&priv->resolver);
 	g_clear_object (&priv->server);
 	g_clear_pointer (&priv->server_context, g_main_context_unref);
-	g_clear_object (&priv->input_stream);
+	g_clear_pointer (&priv->hosts, g_hash_table_unref);
+	g_clear_object (&priv->hosts_trace_file);
+	g_clear_object (&priv->hosts_output_stream);
 	g_clear_object (&priv->trace_file);
 	g_clear_object (&priv->input_stream);
 	g_clear_object (&priv->output_stream);
@@ -545,6 +557,8 @@ server_process_message (UhmServer *self, UhmMessage *message)
 	UhmServerPrivate *priv = self->priv;
 	g_autoptr(GBytes) message_body = NULL;
 	goffset expected_content_length;
+	g_autoptr(GError) error = NULL;
+	const char *location_header = NULL;
 
 	g_assert (priv->next_message != NULL);
 	priv->message_counter++;
@@ -574,6 +588,31 @@ server_process_message (UhmServer *self, UhmMessage *message)
 	uhm_message_set_http_version (message, uhm_message_get_http_version (priv->next_message));
 	uhm_message_set_status (message, uhm_message_get_status (priv->next_message),
 	                        uhm_message_get_reason_phrase (priv->next_message));
+
+	/* Rewrite Location headers to use the uhttpmock server details */
+	location_header = soup_message_headers_get_one (uhm_message_get_response_headers (priv->next_message), "Location");
+	if (location_header) {
+		g_autoptr(GUri) uri = NULL;
+		g_autoptr(GUri) modified_uri = NULL;
+		g_autofree char *uri_str = NULL;
+
+		uri = g_uri_parse (location_header, G_URI_FLAGS_ENCODED, &error);
+		if (uri) {
+			modified_uri = g_uri_build (G_URI_FLAGS_ENCODED,
+				                    g_uri_get_scheme (uri),
+				                    g_uri_get_userinfo (uri),
+				                    g_uri_get_host (uri),
+				                    priv->port,
+				                    g_uri_get_path (uri),
+				                    g_uri_get_query (uri),
+				                    g_uri_get_fragment (uri));
+
+			uri_str = g_uri_to_string (modified_uri);
+			soup_message_headers_replace (uhm_message_get_response_headers (priv->next_message), "Location", uri_str);
+		} else {
+			g_debug ("Failed to rewrite Location header ‘%s’ to use new port", location_header);
+		}
+	}
 	soup_message_headers_foreach (uhm_message_get_response_headers (priv->next_message), header_append_cb, message);
 
 	/* Add debug headers to identify the message and trace file. */
@@ -876,6 +915,9 @@ trace_to_soup_message (const gchar *trace, GUri *base_uri)
 	} else if (strncmp (trace, "HTTP/1.0", strlen ("HTTP/1.0")) == 0) {
 		http_version = SOUP_HTTP_1_0;
 		trace += strlen ("HTTP/1.0");
+	} else if (strncmp (trace, "HTTP/2", strlen ("HTTP/2")) == 0) {
+		http_version = SOUP_HTTP_2_0;
+		trace += strlen ("HTTP/2");
 	} else {
 		g_warning ("Unrecognised HTTP version ‘%s’.", trace);
 		http_version = SOUP_HTTP_1_1;
@@ -917,6 +959,9 @@ trace_to_soup_message (const gchar *trace, GUri *base_uri)
 	} else if (strncmp (trace, "HTTP/1.0", strlen ("HTTP/1.0")) == 0) {
 		http_version = SOUP_HTTP_1_0;
 		trace += strlen ("HTTP/1.0");
+	} else if (strncmp (trace, "HTTP/2", strlen ("HTTP/2")) == 0) {
+		http_version = SOUP_HTTP_2_0;
+		trace += strlen ("HTTP/2");
 	} else {
 		g_warning ("Unrecognised HTTP version ‘%s’.", trace);
 	}
@@ -1144,6 +1189,12 @@ uhm_server_load_trace (UhmServer *self, GFile *trace_file, GCancellable *cancell
 {
 	UhmServerPrivate *priv = self->priv;
 	g_autoptr(GUri) base_uri = NULL;
+	g_autoptr(GError) local_error = NULL;
+	g_autofree char *content = NULL;
+	g_autofree char *trace_path = NULL;
+	g_autofree char *trace_hosts = NULL;
+	g_auto(GStrv) split = NULL;
+	gsize len;
 
 	g_return_if_fail (UHM_IS_SERVER (self));
 	g_return_if_fail (G_IS_FILE (trace_file));
@@ -1153,6 +1204,7 @@ uhm_server_load_trace (UhmServer *self, GFile *trace_file, GCancellable *cancell
 
 	base_uri = build_base_uri (self);
 
+	/* Trace File */
 	priv->trace_file = g_object_ref (trace_file);
 	priv->input_stream = load_file_stream (priv->trace_file, cancellable, error);
 
@@ -1167,10 +1219,34 @@ uhm_server_load_trace (UhmServer *self, GFile *trace_file, GCancellable *cancell
 		if (child_error != NULL) {
 			g_clear_object (&priv->trace_file);
 			g_propagate_error (error, child_error);
+			return;
 		}
 	} else {
 		/* Error. */
 		g_clear_object (&priv->trace_file);
+		return;
+	}
+
+	/* Host file */
+	trace_path = g_file_get_path (trace_file);
+	trace_hosts = g_strconcat (trace_path, ".hosts", NULL);
+	priv->hosts_trace_file = g_file_new_for_path (trace_hosts);
+
+	if (g_file_load_contents (priv->hosts_trace_file, cancellable, &content, &len, NULL, &local_error)) {
+		split = g_strsplit (content, "\n", -1);
+		for (gsize i = 0; split != NULL && split[i] != NULL; i++) {
+			if (*(split[i]) != '\0') {
+				uhm_resolver_add_A (priv->resolver, split[i], uhm_server_get_address (self));
+			}
+		}
+	} else if (g_error_matches (local_error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND)) {
+		/* It's not fatal that this file cannot be loaded as these hosts can be added in code */
+		g_clear_error (&local_error);
+
+	} else {
+		/* Other I/O errors are fatal. */
+		g_propagate_error (error, g_steal_pointer (&local_error));
+		return;
 	}
 }
 
@@ -1545,27 +1621,41 @@ uhm_server_start_trace_full (UhmServer *self, GFile *trace_file, GError **error)
 	}
 	g_return_if_fail (priv->output_stream == NULL);
 
+	if (priv->enable_online == TRUE) {
+		priv->message_counter = 0;
+	    priv->comparison_message = g_byte_array_new ();
+		priv->received_message_state = UNKNOWN;
+	}
+
 	/* Start writing out a trace file if logging is enabled. */
 	if (priv->enable_logging == TRUE) {
 		GFileOutputStream *output_stream;
+		g_autofree char *trace_path = g_file_get_path (trace_file);
+		g_autofree char *trace_hosts = g_strconcat (trace_path, ".hosts", NULL);
+		priv->hosts_trace_file = g_file_new_for_path (trace_hosts);
 
 		output_stream = g_file_replace (trace_file, NULL, FALSE, G_FILE_CREATE_NONE, NULL, &child_error);
 
 		if (child_error != NULL) {
-			gchar *trace_file_path;
-
-			trace_file_path = g_file_get_path (trace_file);
-			g_set_error (error, child_error->domain, child_error->code,
-			             "Error replacing trace file ‘%s’: %s", trace_file_path, child_error->message);
-			g_free (trace_file_path);
-
-			g_error_free (child_error);
-
+			g_propagate_prefixed_error (error, g_steal_pointer (&child_error),
+			             "Error replacing trace file ‘%s’: ", trace_path);
 			return;
 		} else {
 			/* Change state. */
 			priv->output_stream = output_stream;
 		}
+
+		/* Host trace file */
+		output_stream = g_file_replace (priv->hosts_trace_file, NULL, FALSE, G_FILE_CREATE_NONE, NULL, &child_error);
+		if (child_error != NULL) {
+			g_autofree char *hosts_trace_file_path = g_file_get_path (priv->hosts_trace_file);
+			g_propagate_prefixed_error (error, g_steal_pointer (&child_error),
+			             "Error replacing trace hosts file ‘%s’: ", hosts_trace_file_path);
+			return;
+		} else {
+			/* Change state. */
+			priv->hosts_output_stream = output_stream;
+		}
 	}
 
 	/* Start reading from a trace file if online testing is disabled or if we need to compare server responses to the trace file. */
@@ -1630,6 +1720,7 @@ uhm_server_end_trace (UhmServer *self)
 
 	if (priv->enable_logging == TRUE) {
 		g_clear_object (&self->priv->output_stream);
+		g_clear_object (&self->priv->hosts_output_stream);
 	}
 }
 
@@ -1733,6 +1824,8 @@ uhm_server_received_message_chunk (UhmServer *self, const gchar *message_chunk,
 {
 	UhmServerPrivate *priv = self->priv;
 	GError *child_error = NULL;
+	g_autoptr(UhmMessage) online_message = NULL;
+	g_autoptr(GUri) base_uri = NULL;
 
 	g_return_if_fail (UHM_IS_SERVER (self));
 	g_return_if_fail (message_chunk != NULL);
@@ -1794,8 +1887,8 @@ uhm_server_received_message_chunk (UhmServer *self, const gchar *message_chunk,
 
 	/* Append to the trace file. */
 	if (priv->enable_logging == TRUE &&
-	    (g_output_stream_write_all (G_OUTPUT_STREAM (priv->output_stream), message_chunk, message_chunk_length, NULL, NULL, &child_error) == FALSE ||
-	     g_output_stream_write_all (G_OUTPUT_STREAM (priv->output_stream), "\n", 1, NULL, NULL, &child_error) == FALSE)) {
+	    (!g_output_stream_write_all (G_OUTPUT_STREAM (priv->output_stream), message_chunk, message_chunk_length, NULL, NULL, &child_error) ||
+	     !g_output_stream_write_all (G_OUTPUT_STREAM (priv->output_stream), "\n", 1, NULL, NULL, &child_error))) {
 		gchar *trace_file_path = g_file_get_path (priv->trace_file);
 		g_set_error (error, child_error->domain, child_error->code,
 		             "Error appending to log file ‘%s’: %s", trace_file_path, child_error->message);
@@ -1806,8 +1899,8 @@ uhm_server_received_message_chunk (UhmServer *self, const gchar *message_chunk,
 		return;
 	}
 
-	/* Or compare to the existing trace file. */
-	if (priv->enable_logging == FALSE && priv->enable_online == TRUE) {
+	/* Update comparison message */
+	if (priv->enable_online == TRUE) {
 		/* Build up the message to compare. We explicitly don't escape nul bytes, because we want the trace
 		 * files to be (pretty much) ASCII. File uploads are handled by zero-extending the responses according
 		 * to the traced Content-Length. */
@@ -1815,36 +1908,47 @@ uhm_server_received_message_chunk (UhmServer *self, const gchar *message_chunk,
 		g_byte_array_append (priv->comparison_message, (const guint8 *) "\n", 1);
 
 		if (priv->received_message_state == RESPONSE_TERMINATOR) {
-			/* Received the last chunk of the response, so compare the message from the trace file and that from online. */
-			UhmMessage *online_message;
-			g_autoptr(GUri) base_uri = NULL;
-
 			/* End of a message. */
 			base_uri = build_base_uri (self);
 			online_message = trace_to_soup_message ((const gchar *) priv->comparison_message->data, base_uri);
 
 			g_byte_array_set_size (priv->comparison_message, 0);
 			priv->received_message_state = UNKNOWN;
+		}
+	}
 
-			g_assert (priv->next_message != NULL);
+	/* Append to the hosts file */
+	if (online_message != NULL && priv->enable_online == TRUE && priv->enable_logging == TRUE) {
+		const char *host = soup_message_headers_get_one (uhm_message_get_request_headers (online_message), "Soup-Host");
 
-			/* Compare the message from the server with the message in the log file. */
-			if (compare_incoming_message (self, online_message, priv->next_message) != 0) {
-				gchar *next_uri, *actual_uri;
+		if (!g_output_stream_write_all (G_OUTPUT_STREAM (priv->hosts_output_stream), host, strlen (host), NULL, NULL, &child_error)  ||
+		    !g_output_stream_write_all (G_OUTPUT_STREAM (priv->hosts_output_stream), "\n", 1, NULL, NULL, &child_error)) {
+			g_autofree gchar *hosts_trace_file_path = g_file_get_path (priv->hosts_trace_file);
+			g_warning ("Error appending to host log file ‘%s’: %s", hosts_trace_file_path, child_error->message);
+		}
 
-				next_uri = uri_get_path_query (uhm_message_get_uri (priv->next_message));
-				actual_uri = uri_get_path_query (uhm_message_get_uri (online_message));
-				g_set_error (error, UHM_SERVER_ERROR, UHM_SERVER_ERROR_MESSAGE_MISMATCH,
-				             "Expected URI ‘%s’, but got ‘%s’.", next_uri, actual_uri);
-				g_free (actual_uri);
-				g_free (next_uri);
+		if (host != NULL)
+			g_hash_table_add (priv->hosts, g_strdup (host));
+	}
 
-				g_object_unref (online_message);
+	/* Or compare to the existing trace file. */
+	if (online_message != NULL && priv->enable_logging == FALSE && priv->enable_online == TRUE) {
+		/* Received the last chunk of the response, so compare the message from the trace file and that from online. */
+		g_assert (priv->next_message != NULL);
 
-				return;
-			}
+		/* Compare the message from the server with the message in the log file. */
+		if (compare_incoming_message (self, online_message, priv->next_message) != 0) {
+			gchar *next_uri, *actual_uri;
+
+			next_uri = uri_get_path_query (uhm_message_get_uri (priv->next_message));
+			actual_uri = uri_get_path_query (uhm_message_get_uri (online_message));
+			g_set_error (error, UHM_SERVER_ERROR, UHM_SERVER_ERROR_MESSAGE_MISMATCH,
+			             "Expected URI ‘%s’, but got ‘%s’.", next_uri, actual_uri);
+			g_free (actual_uri);
+			g_free (next_uri);
 
 			g_object_unref (online_message);
+			return;
 		}
 	}
 }
